==> splitp-0.1.1-py3-none-any.whl: splitp/tree_helper_functions.py <==
import numpy as np
import copy as cp
import pandas as pd
import itertools
import collections
from itertools import permutations

def get_balance(s, asTuple=False):
    """Returns a string formatted 'X|X' which describes the balance of a given split string"""
    s = s.split("|")
    if not asTuple:
        return str(len(s[0])) + '|' + str(len(s[1]))
    else:
        return (len(s[0]), len(s[1]))

def frob_norm(m):
    """Calculates the Frobenius Norm for a given numpy array"""
    norm = 0
    for i in np.nditer(m):
        norm += i ** 2
    norm = np.sqrt(norm)
    return norm

def make_substitution_matrix(subs_prob, k):
    matrix = []
    for i in range(k):
        matrix.append([1-subs_prob if j==i else subs_prob/(k-1) for j in range(k)])
    return matrix

def generate_all_splits(num_taxa, trivial=True, onlyTrivial=False, onlyBalance=-1):
    """Generates all splits as string-representations

    Args:
        num_taxa: The number of taxa on the tree.
        trivial: Whether or not to calculate trivial splits, default True.
        onlyTrivial: Whether to ONLY create trivial splits, default False.

    Returns:
        A list of string-representations of splits (using '|'-notation)
    """
    s = [str(i) for i in range(num_taxa)]
    s.append('|')
    perms = [''.join(p) for p in permutations(s)]
    perms = set(perms)
    aList = []
    complete_set = set()
    for p in perms:
        halves = p.split('|')
        limit = 0
        if not trivial: limit = 1
        if onlyBalance == -1:
            criterion = (len(halves[0]) > limit) and (len(halves[1]) > limit)
        else:
            criterion = (len(halves[0]) == onlyBalance) or (len(halves[1]) == onlyBalance)
        if criterion:
            a = ''.join(sorted(halves[0]))
            b = ''.join(sorted(halves[1]))
            if a not in aList and b not in aList:
                if '0' in a:
                    complete_set.add(a + '|' + b)
                else:
                    complete_set.add(b + '|' + a)
                aList.append(a)
                aList.append(b)

    if onlyTrivial:
        new_set = complete_set.copy()
        for s in complete_set:
            halves = s.split('|')
            if len(halves[0]) != 1 and len(halves[1]) != 1:
                new_set.remove(s)
        return list(new_set)

    return sorted(list(complete_set))


###
# Functions for reading in a sequence alignment

def read_alignment_from_file(pathToFile, check=True):
    file = open(pathToFile, 'r')
    alignment = collections.OrderedDict()  # Keeps insertion order
    currentSeq = ""
    for line in file:
        if '>' in line:
            currentSeq = line.replace('>', '').replace('\n', '')
        else:
            if currentSeq in alignment:
                alignment[currentSeq] += line.replace('\n', '')
            else:
                alignment[currentSeq] = line.replace('\n', '')
    if check:
        return alignment if __valid_alignment(alignment) else None
    else:
        return alignment


def __valid_alignment(alignmentDict):
    ordered = isinstance(alignmentDict, collections.OrderedDict)  # Must be ordered to avoid incorrect patterns
    sameLength = len(set(len(value) for key, value in alignmentDict.items())) == 1  # Must be same length
    return sameLength and ordered


def get_pattern_counts(alignment, asNumpyArray=False):
    patterns = {}
    sequences = list(alignment.values())
    sequenceLength = len(sequences[0])
    usableSequenceLength = 0
    for i in range(sequenceLength):
        patternAtSitei = ''.join(s[i] for s in sequences).upper()
        if all(c in ('A', 'C', 'G', 'T') for c in patternAtSitei):  # Need to handle U (let U = T)
            usableSequenceLength += 1
            if patternAtSitei in patterns:
                patterns[patternAtSitei] += 1
            else:
                patterns[patternAtSitei] = 1

    if asNumpyArray:
        patterns = np.array([[key, val] for key, val in patterns.items()])

        file = open("output/patternCounts", 'w')
    for p in patterns:
        file.write(str(p))
    file.close()

    return patterns, usableSequenceLength


def pattern_counts_to_probs(patterns, seqLen):
    newCounts = [float(float(count) / seqLen) for count in patterns[:, 1]]
    patternList = patterns[:, 0]
    return np.array([patternList, newCounts]).transpose()


def pattern_probs_from_alignment(pathToFile, check=True):
    alignment = read_alignment_from_file(pathToFile, check)
    counts, sequenceLength = get_pattern_counts(alignment, True)
    probs = pattern_counts_to_probs(counts, sequenceLength)
    probs = pd.DataFrame(probs, index=probs[:, 0])
    probs[[1]] = probs[[1]].astype("float")
    return probs, len(alignment), sequenceLength


def scores_to_weights(scores):
    scores = [s if s != 0 else 10 ** (-100) for s in scores]
    inverses = [1 / s for s in scores]
    weights = [i / sum(inverses) for i in inverses]
    return weights


####

# Function for scaling the S part of the hard-coded H matrix
def scaled_h_matrix(_lambda):
    if _lambda == None:
        return None, None
    if _lambda == "identity":
        return np.matrix([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [1, 1, 1, 1]]), "identity"
    H = np.matrix([[1, -1], [1, 1]])
    H = np.kron(H, H)
    S = np.empty((0, 4))
    for row in H:
        if len(set(np.asarray(row)[0])) != 1:
            S = np.append(S, row, axis=0)
    S = _lambda * S
    H = np.append(S, [[1, 1, 1, 1]], axis=0)
    return (H, _lambda)

==> splitp-0.1.5-py3-none-any.whl: splitp/tree_helper_functions.py <==
import numpy as np
import pandas as pd
import collections
from scipy.sparse import issparse

def balanced_newick_tree(num_taxa):
    if num_taxa%2 != 0:
        raise ValueError("There is no balanced tree on {num_taxa} taxa. Please specify an even number.")
    from math import floor
    def _balanced_newick_subtree(nt, left=False):
        if nt == 2:
            return "(_,_)"
        elif nt==3:
            return "((_,_),_)" if left else "(_,(_,_))"
        else:
            if nt%2==0:
                return f"({_balanced_newick_subtree(nt/2, True)},{_balanced_newick_subtree(nt/2)})"
            else: 
                return f"({_balanced_newick_subtree(floor(nt/2) + int(left), True)},{_balanced_newick_subtree(floor(nt/2) + int(not left))})"
    newick_string = f"({_balanced_newick_subtree(num_taxa/2, True)},{_balanced_newick_subtree(num_taxa/2)});"
    for i in range(0, num_taxa):
        newick_string = newick_string.replace('_', str(i), 1)
    return newick_string

def get_balance(s, asTuple=False):
    """Returns a string formatted 'X|X' which describes the balance of a given split string"""
    s = s.split("|")
    if not asTuple:
        return str(len(s[0])) + '|' + str(len(s[1]))
    else:
        return (len(s[0]), len(s[1]))

def is_sparse(matrix):
    return issparse(matrix)

def frob_norm(matrix, data_table=None):
    """Calculates the Frobenius Norm for a given matrix"""
    if data_table is not None:
        return sum(val**2 for _, val in data_table.itertuples(index=False))**(1/2)
    if is_sparse(matrix):
        return sum(matrix[i,j]**2 for i, j in zip(*matrix.nonzero()))**(1/2)
    else:
        return np.sqrt(sum(val**2 for val in np.nditer(matrix)))
        
def make_substitution_matrix(subs_prob, k):
    matrix = []
    for i in range(k):
        matrix.append([1-subs_prob if j==i else subs_prob/(k-1) for j in range(k)])
    return matrix

def all_splits(num_taxa, trivial=False, only_balance=None, randomise=False):
    """Generates all splits as string-representations

    Args:
        num_taxa: The number of taxa on the tree.
        trivial: Whether or not to calculate trivial splits, default True.
        only_trivial: Whether to ONLY create trivial splits, default False.

    Returns:
        A list of string-representations of splits (using '|'-notation)
    """
    k = only_balance
    n = num_taxa
    taxa_string = "".join(np.base_repr(i, base=n) for i in range(n))
    r = 0 if trivial else 1
    loop_over = range(r, 2**(n-1) - r)
    if randomise:
        import random
        loop_over = [i for i in loop_over]
        random.shuffle(loop_over)
    for i in loop_over:
        template = format(i, f'0{n}b')
        if not only_balance or sum(int(b) for b in template) in [only_balance, n-only_balance]:
            if r < sum(int(b) for b in template) < n-r:
                left  = ""
                right = ""
                for t, b in zip(taxa_string, template):
                    if b == '0':
                        left += t
                    else:
                        right += t
                if '0' in right:
                    left, right = right, left
                yield f'{left}|{right}'
###
# Functions for reading in a sequence alignment

def read_alignment_from_file(pathToFile, check=True):
    file = open(pathToFile, 'r')
    alignment = collections.OrderedDict()  # Keeps insertion order
    currentSeq = ""
    for line in file:
        if '>' in line:
            currentSeq = line.replace('>', '').replace('\n', '')
        else:
            if currentSeq in alignment:
                alignment[currentSeq] += line.replace('\n', '')
            else:
                alignment[currentSeq] = line.replace('\n', '')
    if check:
        return alignment if __valid_alignment(alignment) else None
    else:
        return alignment


def __valid_alignment(alignmentDict):
    ordered = isinstance(alignmentDict, collections.OrderedDict)  # Must be ordered to avoid incorrect patterns
    sameLength = len(set(len(value) for key, value in alignmentDict.items())) == 1  # Must be same length
    return sameLength and ordered


def get_pattern_counts(alignment, asNumpyArray=False):
    patterns = {}
    sequences = list(alignment.values())
    sequenceLength = len(sequences[0])
    usableSequenceLength = 0
    for i in range(sequenceLength):
        patternAtSitei = ''.join(s[i] for s in sequences).upper()
        if all(c in ('A', 'C', 'G', 'T') for c in patternAtSitei):  # Need to handle U (let U = T)
            usableSequenceLength += 1
            if patternAtSitei in patterns:
                patterns[patternAtSitei] += 1
            else:
                patterns[patternAtSitei] = 1

    if asNumpyArray:
        patterns = np.array([[key, val] for key, val in patterns.items()])

    return patterns, usableSequenceLength


def pattern_counts_to_probs(patterns, seqLen):
    newCounts = [float(float(count) / seqLen) for count in patterns[:, 1]]
    patternList = patterns[:, 0]
    return np.array([patternList, newCounts]).transpose()


def pattern_probs_from_alignment(pathToFile, check=True):
    alignment = read_alignment_from_file(pathToFile, check)
    counts, sequenceLength = get_pattern_counts(alignment, True)
    probs = pattern_counts_to_probs(counts, sequenceLength)
    probs = pd.DataFrame(probs, index=probs[:, 0])
    probs[[1]] = probs[[1]].astype("float")
    return probs, len(alignment), sequenceLength


def scores_to_weights(scores):
    scores = [s if s != 0 else 10 ** (-100) for s in scores]
    inverses = [1 / s for s in scores]
    weights = [i / sum(inverses) for i in inverses]
    return weights


####

# Function for scaling the S part of the hard-coded H matrix
def scaled_h_matrix(_lambda):
    if _lambda == None:
        return None, None
    if _lambda == "identity":
        return np.matrix([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [1, 1, 1, 1]]), "identity"
    H = np.matrix([[1, -1], [1, 1]])
    H = np.kron(H, H)
    S = np.empty((0, 4))
    for row in H:
        if len(set(np.asarray(row)[0])) != 1:
            S = np.append(S, row, axis=0)
    S = _lambda * S
    H = np.append(S, [[1, 1, 1, 1]], axis=0)
    return (H, _lambda)
